# Dummy class for packages that have no MPI
class MPIDummy(object):
    def __init__(self):
        pass

    def Get_rank(self):
        return 0

    def Get_size(self):
        return 1

    def barrier(self):
        pass

    def send(self, lnlike0, dest=1, tag=55):
        pass

    def recv(self, source=1, tag=55):
        pass

    def Iprobe(self, source=1, tag=55):
        pass

    def scatter(self, sendobj, **kwargs):
        if sendobj is not None:
            return sendobj[0]
        return None

    def bcast(self, obj, **kwargs):
        return obj

    def gather(self, sendobj, **kwargs):
        return [sendobj]


# Global object representing no MPI:
COMM_WORLD = MPIDummy()



"""
Integration tests to make sure the sampler at least runs. Better than nothing...
"""
import shutil
from unittest import TestCase

import numpy as np
from mpi4py import MPI

from PTMCMCSampler import PTMCMCSampler
from PTMCMCSampler import nompi4py as MPIDUMMY


class GaussianLikelihood(object):
    def __init__(self, ndim=2, pmin=-10, pmax=10):

        self.a = np.ones(ndim) * pmin
        self.b = np.ones(ndim) * pmax

        # get means
        self.mu = np.random.uniform(pmin, pmax, ndim)

        # ... and a positive definite, non-trivial covariance matrix.
        cov = 0.5 - np.random.rand(ndim**2).reshape((ndim, ndim))
        cov = np.triu(cov)
        cov += cov.T - np.diag(cov.diagonal())
        self.cov = np.dot(cov, cov)

        # Invert the covariance matrix first.
        self.icov = np.linalg.inv(self.cov)

    def lnlikefn(self, x):
        diff = x - self.mu
        return -np.dot(diff, np.dot(self.icov, diff)) / 2.0

    def lnpriorfn(self, x):

        if np.all(self.a <= x) and np.all(self.b >= x):
            return 0.0
        else:
            return -np.inf


class UniformJump(object):
    def __init__(self, pmin, pmax):
        """Draw random parameters from pmin, pmax"""
        self.pmin = pmin
        self.pmax = pmax

    def jump(self, x, it, beta):
        """
        Function prototype must read in parameter vector x,
        sampler iteration number it, and inverse temperature beta
        """

        # log of forward-backward jump probability
        lqxy = 0

        # uniformly drawm parameters
        q = np.random.uniform(self.pmin, self.pmax, len(x))

        return q, lqxy


class TestSimpleSampler(TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("chains")

    def setUp(self) -> None:
        self.comm = MPI.COMM_WORLD

    def test_simple(self):
        # ## Setup Gaussian model class
        ndim = 20
        pmin, pmax = 0.0, 10.0
        glo = GaussianLikelihood(ndim=ndim, pmin=pmin, pmax=pmax)

        # ## Setup sampler
        # Set the start position and the covariance
        p0 = np.random.uniform(pmin, pmax, ndim)
        cov = np.eye(ndim) * 0.1**2

        sampler = PTMCMCSampler.PTSampler(
            ndim,
            glo.lnlikefn,
            glo.lnpriorfn,
            np.copy(cov),
            outDir="./chains",
            comm=self.comm,
        )

        # add to jump proposal cycle
        ujump = UniformJump(pmin, pmax)
        sampler.addProposalToCycle(ujump.jump, 5)

        sampler.sample(p0, 10000, burn=500, thin=1, covUpdate=500, SCAMweight=20, AMweight=20, DEweight=20)


class TestSimpleSamplerNoMPI(TestSimpleSampler):
    def setUp(self) -> None:
        self.comm = MPIDUMMY.COMM_WORLD

    def test_simple(self):
        return super().test_simple()



"""
Implementation of the No-U-Turn-Sampler. Code follows algorithm 6 from the NUTS
paper (Hoffman & Gelman, 2011)

reference: arXiv:1111.4246
"The No-U-Turn Sampler: Adaptively Setting Path Lengths in Hamiltonian Monte
Carlo", Matthew D. Hoffman & Andrew Gelman

Rutger van Haasteren
"""


import os
import sys

import numpy as np
import scipy.linalg as sl


class GradientJump(object):
    """Class for jumps using gradient information"""

    def __init__(self, loglik_grad, logprior_grad, mm_inv, nburn=100):
        """Initialize the HMC class

        :loglik_grad:   Log-likelihood and gradient function
        :logprior_grad: Log-prior and gradient function
        :mm_inv:        Inverse of the mass matrix (covariance matrix)
        :nburn:         Number of burn-in steps
        """

        self._loglik_grad = loglik_grad  # Log-likelihood & gradient
        self._logprior_grad = logprior_grad  # Log-prior & gradient
        self.mm_inv = mm_inv  # Inverse mass-matrix
        self.nburn = nburn  # Nr. of burn-in steps
        self.ndim = len(self.mm_inv)  # Number of dimensions
        self.set_cf()  # Whitening matrices

        self.name = "GradientJUMP"

        self.epsilon = None  # Step-size
        self.beta = 1.0  # Inverse temperature
        self.iter = 0.0  # Number of gradient jumps

        print("WARNING: GradientJumps not yet adaptive. Choose cov wisely!")

    @property
    def __name__(self):
        return self.name

    def set_cf(self):
        """Update the Cholesky factor of the inverse mass matrix"""
        self.cov_cf = sl.cholesky(self.mm_inv, lower=True)
        self.cov_cfi = sl.solve_triangular(self.cov_cf, np.eye(len(self.cov_cf)), trans=0, lower=True)

    def update_cf(self):
        """Update the Cholesky factor of the inverse mass matrix

        NOTE: this function is different from the one in GradientJump!
        """
        # Since we are adaptively tuning the step size epsilon, we should at
        # least keep the determinant of this guy equal to what it was before.
        new_cov_cf = sl.cholesky(self.mm_inv, lower=True)

        ldet_old = np.sum(np.log(np.diag(self.cov_cf)))
        ldet_new = np.sum(np.log(np.diag(new_cov_cf)))

        self.cov_cf = np.exp((ldet_old - ldet_new) / self.ndim) * new_cov_cf
        self.cov_cfi = sl.solve_triangular(self.cov_cf, np.eye(len(self.cov_cf)), trans=0, lower=True)

    def func_grad(self, x):
        """Log-prob and gradient, corrected for temperature"""
        ll, ll_grad = self._loglik_grad(x)
        lp, lp_grad = self._logprior_grad(x)

        return self.beta * ll + lp, self.beta * ll_grad + lp_grad

    def forward(self, x):
        """Coordinate transformation to whitened parameters x->q"""
        return np.dot(self.cov_cfi.T, x)

    def backward(self, q):
        """Coordinate transformation from whitened parameters q->x"""
        return np.dot(self.cov_cf.T, q)

    def func_grad_white(self, q):
        """Whitened version of func_grad"""
        x = self.backward(q)
        fv, fg = self.func_grad(x)
        return fv, np.dot(self.cov_cf, fg)

    def draw_momenta(self):
        """Draw new momentum variables"""
        return np.random.randn(len(self.mm_inv))

    def loghamiltonian(self, logl, r):
        """Value of the Hamiltonian, given a position and momentum value"""
        try:
            return logl - 0.5 * np.dot(r, r)
        except ValueError:
            return np.nan

    def posmom_inprod(self, theta, r):
        try:
            return np.dot(theta, r)
        except ValueError:
            return np.nan

    # With the following definitions, we should be able to get rid of this
    # awkward coordinate transformation. Why does it not work?
    """
    def forward(self, x):
        #return np.dot(self.cov_cfi.T, x)
        return x

    def backward(self, q):
        #return np.dot(self.cov_cf.T, q)
        return q

    def func_grad_white(self, q):
        # We would be able to get rid of this function
        #x = self.backward(q)
        #fv, fg = self.func_grad(x)
        #return fv, np.dot(self.cov_cf, fg)
        return self.func_grad(q)

    def draw_momenta(self):
        #return np.random.randn(len(self.mm_inv))
        return np.dot(self.cov_cfi.T, np.random.randn(self.ndim))

    def loghamiltonian(self, logl, r):
        try:
            #return logl-0.5*np.dot(r, r)
            newr = np.dot(self.cov_cfi.T, r)
            return logl-0.5*np.dot(newr, newr)
        except ValueError as err:
            return np.nan

    def posmom_inprod(self, theta, r):
        try:
            newr = np.dot(self.cov_cfi.T, r)
            newtheta = np.dot(self.cov_cfi.T, theta)
            return np.dot(newtheta, newr)
            #return np.dot(theta, r)
        except ValueError as err:
            return np.nan
    """

    def leapfrog(self, theta, r, grad, epsilon):
        """Perfom a leapfrog jump in the Hamiltonian space

        :theta:     Initial parameter position
        :r:         Initial momentum
        :grad:      Initial gradient
        :epsilon:   Step size

        output
        thetaprime: new parameter position
        rprime:     new momentum
        gradprime:  new gradient
        logpprime:  new log-probability
        """

        rprime = r + 0.5 * epsilon * grad  # half step in r
        thetaprime = theta + epsilon * rprime  # step in theta
        logpprime, gradprime = self.func_grad_white(thetaprime)  # compute gradient
        rprime = rprime + 0.5 * epsilon * gradprime  # half step in r

        return thetaprime, rprime, gradprime, logpprime

    def __call__(self, x, iter, beta):
        """Take one HMC trajectory step"""

        self.iter += 1

        if self.__name__ == "GradientJUMP":
            raise NotImplementedError("GradientJump is an abstract base class!")
        else:
            return x, 0.0


class MALAJump(GradientJump):
    """MALA Jump"""

    def __init__(self, loglik_grad, logprior_grad, mm_inv, nburn=100):
        """Initialize the MALA Jump"""
        super(MALAJump, self).__init__(loglik_grad, logprior_grad, mm_inv, nburn=nburn)

        self.name = "MALAJump"
        self.cd = 2.4 / np.sqrt(self.ndim)
        self.set_eigvecs()

    def set_eigvecs(self):
        """Set the eigenvectors of the mass matrix"""
        # Since we have whitened the parameter space, the decomposition is a
        # simple identity matrix
        self._u = np.eye(self.ndim)
        self._s = np.ones(self.ndim)

    def __call__(self, x, iter, beta):
        """Take one MALA step"""
        super(MALAJump, self).__call__(x, iter, beta)

        x = np.atleast_1d(x)
        if len(np.shape(x)) > 1:
            raise ValueError("x is expected to be a 1-D array")

        self.beta = beta

        # Update the mass matrix when in burn-in stage
        # Currently, there is a problem with adjusting the mass matrix.
        # Stepsize and mass matrix need to be tuned together?
        # if iter <= self.nburn:
        #    self.update_cf()
        #    self.set_eigvecs()

        # Initial starting position
        q0 = self.forward(x)
        logp, grad0 = self.func_grad_white(q0)

        # Choose an eigenvector to jump in, and the size
        i = np.random.randint(0, self.ndim)
        vec = self._u[i, :]
        val = self._s[i]
        dist = np.random.randn()

        # Do the leapfrog
        mq0 = q0 + 0.5 * vec * self.cd**2 * np.dot(vec, grad0) / 2 / val
        q1 = mq0 + dist * vec * self.cd / np.sqrt(val)
        logp1, grad1 = self.func_grad_white(q1)
        mq1 = q1 + 0.5 * vec * self.cd**2 * np.dot(vec, grad1) / 2 / val

        qxy = 0.5 * (np.sum((mq0 - q1) ** 2 / val) - np.sum((mq1 - q0) ** 2 / val))

        return self.backward(q1), qxy


class HMCJump(GradientJump):
    """Hamiltonian Monte Carlo Jump"""

    def __init__(self, loglik_grad, logprior_grad, mm_inv, nburn=100, stepsize=0.1, nminsteps=10, nmaxsteps=300):
        """Initialize the MALA Jump"""
        super(HMCJump, self).__init__(loglik_grad, logprior_grad, mm_inv, nburn=nburn)

        self.name = "HMCJump"
        self.epsilon = stepsize
        self.nminsteps = nminsteps
        self.nmaxsteps = nmaxsteps

    def __call__(self, x, iter, beta):
        """Take one HMC step"""
        super(HMCJump, self).__call__(x, iter, beta)

        x = np.atleast_1d(x)
        if len(np.shape(x)) > 1:
            raise ValueError("x is expected to be a 1-D array")

        # Set the temperature
        self.beta = beta

        # Update the mass matrix when in burn-in stage
        # Currently, there is a problem with adjusting the mass matrix.
        # Stepsize and mass matrix need to be tuned together?
        # if iter <= self.nburn:
        #    self.update_cf()

        # Initial starting position
        q0 = self.forward(x)
        qxy = 0
        logp0, grad0 = self.func_grad_white(q0)

        # Draw new momentum variables
        p0 = self.draw_momenta()
        joint0 = self.loghamiltonian(logp0, p0)

        # Initialize the state
        nsteps = np.random.randint(self.nminsteps, self.nmaxsteps)
        p, q, grad = np.copy(p0), np.copy(q0), np.copy(grad0)

        for ii in range(nsteps):
            q1, p1, grad1, logp1 = self.leapfrog(q, p, grad, self.epsilon)
            joint1 = self.loghamiltonian(logp1, p1)
            p, q, grad = np.copy(p1), np.copy(q1), np.copy(grad1)

            if (joint1 - 1000.0) < joint0:
                # We are super inaccurate, so break the trajectory
                break

        qxy = joint1 - joint0

        return self.backward(q), qxy


class Trajectory(object):
    """Keep track of trajectories in the NUTS jump"""

    def __init__(self, ndim, bufsize=1000):
        """Initialize the trajectory object"""
        self.ndim = ndim
        self.bufadd = bufsize
        self.bufsize_plus = bufsize
        self.bufsize_minus = bufsize
        self.trajlen_plus = 0
        self.trajlen_minus = 0

        self.trajbuf_plus = np.zeros((self.bufsize_plus, self.ndim))
        self.trajind_plus = np.zeros(self.bufsize_plus)
        self.trajbuf_minus = np.zeros((self.bufsize_minus, self.ndim))
        self.trajind_minus = np.zeros(self.bufsize_minus)

    def increase_buf(self, which="plus"):
        """Increase the buffer on the positive or the negative side"""
        addbuf = np.zeros((self.bufadd, self.ndim))
        addind = np.zeros(self.bufadd)

        if which == "plus":
            self.trajbuf_plus = np.append(self.trajbuf_plus, addbuf, axis=0)
            self.trajind_plus = np.append(self.trajind_plus, addind)
            self.bufsize_plus += self.bufadd
        elif which == "minus":
            self.trajbuf_minus = np.append(self.trajbuf_minus, addbuf, axis=0)
            self.trajind_minus = np.append(self.trajind_minus, addind)
            self.bufsize_minus += self.bufadd

    def reset(self):
        """Reset the trajectory object"""
        self.trajlen_plus = 0
        self.trajlen_minus = 0

    def add_sample(self, theta, ind, which="plus"):
        """Add a sample on the positive or the negative branch"""
        if which == "plus":
            if self.trajlen_plus >= self.bufsize_plus:
                self.increase_buf(which="plus")

            self.trajbuf_plus[self.trajlen_plus, :] = theta
            self.trajind_plus[self.trajlen_plus] = ind
            self.trajlen_plus += 1
        elif which == "minus":
            if self.trajlen_minus >= self.bufsize_minus:
                self.increase_buf(which="minus")

            self.trajbuf_minus[self.trajlen_minus, :] = theta
            self.trajind_minus[self.trajlen_minus] = ind
            self.trajlen_minus += 1

    def length(self):
        """Function that returns the current trajectory length"""
        return self.trajlen_plus + self.trajlen_minus

    def get_trajectory(self, which="both"):
        if which == "both":
            return (
                np.append(
                    self.trajbuf_minus[: self.trajlen_minus : -1, :], self.trajbuf_plus[: self.trajlen_plus, :], axis=0
                ),
                np.append(self.trajind_minus[: self.trajlen_minus : -1], self.trajind_plus[: self.trajlen_plus]),
            )
        elif which == "plus":
            return self.trajbuf_plus[: self.trajlen_plus], self.trajind_plus[: self.trajlen_plus]
        elif which == "minus":
            return self.trajbuf_minus[: self.trajlen_minus], self.trajind_minus[: self.trajlen_minus]

    def get_used_trajectory(self, ind):
        """For index ind, get the trajectory that gets us there"""
        tiplus = self.trajind_plus[: self.trajlen_plus]
        timinus = self.trajind_minus[: self.trajlen_minus]

        if ind in tiplus:
            index = np.where(tiplus == ind)[0][0] + 1
            return self.trajbuf_plus[:index, :]
        elif ind in timinus:
            index = np.where(timinus == ind)[0][0] + 1
            return np.append(self.trajbuf_plus[:1, :], self.trajbuf_minus[:index, :], axis=0)
        else:
            raise ValueError("Index not found")


class NUTSJump(GradientJump):
    """Class for No-U-Turn-Sampling Hamiltonian Monte Carlo jumps"""

    def __init__(
        self,
        loglik_grad,
        logprior_grad,
        mm_inv,
        nburn=100,
        trajectoryDir=None,
        write_burnin=False,
        force_trajlen=None,
        force_epsilon=None,
        delta=0.6,
    ):
        """Initialize the HMC class

        :loglik_grad:   Log-likelihood and gradient function
        :logprior_grad: Log-prior and gradient function
        :mm_inv:        Inverse of the mass matrix (covariance matrix)
        :nburn:         Number of burn-in steps
        :trajectoryDir: Output directory for full trajectories (for debugging)
        :write_burnin:  Whether we are writing the burn-in trajectories
        """
        super(NUTSJump, self).__init__(loglik_grad, logprior_grad, mm_inv, nburn=nburn)

        self.trajectoryDir = trajectoryDir  # Trajectory directory
        self.write_burnin = write_burnin  # Write burnin trajectories?

        self.name = "NUTSJUMP"

        self.delta = delta  # Target acceptance rate

        self.traj = Trajectory(self.ndim, bufsize=1000)  # Trajectory buffer

        # Parameters for the dual averaging (for tuning epsilon)
        self.gamma = 0.05
        self.t0 = 10
        self.kappa = 0.75
        self.mu = None
        self.epsilonbar = 1.0
        self.Hbar = 0

        # Parameters to force the trajectories to be pre-set
        self.force_trajlen = force_trajlen
        self.force_epsilon = force_epsilon
        if self.force_epsilon is not None:
            self.epsilonbar = self.force_epsilon

        # Create the trajectory directory, if it does not exist yet
        if self.trajectoryDir is not None:
            if os.path.isfile(trajectoryDir):
                raise IOError("Not a directory: {0}".format(trajectoryDir))
            elif not os.path.isdir(trajectoryDir):
                os.mkdir(trajectoryDir)

    def find_reasonable_epsilon(self, theta0, grad0, logp0):
        """Heuristic for choosing an initial value of epsilon"""

        epsilon = 1.0
        r0 = self.draw_momenta()

        # Figure out what direction we should be moving epsilon.
        _, rprime, gradprime, logpprime = self.leapfrog(theta0, r0, grad0, epsilon)

        # Make sure the step is not too large, so that the likelihood is finite
        # and also stays inside the prior domain (if any)
        k = 1.0
        while np.isinf(logpprime) or np.isinf(gradprime).any():
            k *= 0.5
            _, rprime, _, logpprime = self.leapfrog(theta0, r0, grad0, epsilon * k)

        epsilon = 0.5 * k * epsilon

        acceptprob = np.exp(self.loghamiltonian(logpprime, rprime) - self.loghamiltonian(logp0, r0))

        a = 2.0 * float((acceptprob > 0.5)) - 1.0
        # Keep moving epsilon in that direction until acceptprob crosses 0.5.
        while (acceptprob**a) > (2.0 ** (-a)):
            epsilon = epsilon * (2.0**a)
            _, rprime, _, logpprime = self.leapfrog(theta0, r0, grad0, epsilon)

            acceptprob = np.exp(self.loghamiltonian(logpprime, rprime) - self.loghamiltonian(logp0, r0))

        return epsilon

    def stop_criterion(self, thetaminus, thetaplus, rminus, rplus, force_trajlen, index):
        """Compute the stop condition in the main loop
        dot(dtheta, rminus) >= 0 & dot(dtheta, rplus >= 0)

        INPUTS
        ------
        thetaminus, thetaplus: ndarray[float, ndim=1]
            under and above position
        rminus, rplus: ndarray[float, ndim=1]
            under and above momentum

        OUTPUTS
        -------
        criterion: bool
            return if the condition is valid
        """
        dtheta = thetaplus - thetaminus
        inprod_min = self.posmom_inprod(dtheta, rminus)
        inprod_plus = self.posmom_inprod(dtheta, rplus)

        # orig = (np.dot(dtheta, rminus.T) >= 0) & (np.dot(dtheta, rplus.T) >= 0)
        orig = (inprod_min >= 0) & (inprod_plus >= 0)

        if force_trajlen is not None:
            cont = index < force_trajlen
        else:
            cont = orig

        return cont

    def build_tree(self, theta, r, grad, logu, v, j, epsilon, joint0, ind, traj, force_trajlen):
        """The main recursion tree. Literally from Hoffman and Gelman (2011)."""

        if j == 0:
            # Base case: Take a single leapfrog step in the direction v.
            thetaprime, rprime, gradprime, logpprime = self.leapfrog(theta, r, grad, v * epsilon)
            joint = self.loghamiltonian(logpprime, rprime)

            # Is the new point in the slice of the slice sampling step?
            nprime = int(logu < joint)
            # Is the simulation very inaccurate?
            sprime = int((logu - 1000.0) < joint)

            # Set the return values---minus=plus for all things here, since the
            # "tree" is of depth 0.
            thetaminus = thetaprime[:]
            thetaplus = thetaprime[:]
            rminus = rprime[:]
            rplus = rprime[:]
            gradminus = gradprime[:]
            gradplus = gradprime[:]

            # Compute the acceptance probability.
            alphaprime = min(1.0, np.exp(joint - joint0))
            nalphaprime = 1

            if v == 1 and traj is not None:
                ind_plus, ind_minus = ind + 1, ind
                traj.add_sample(thetaprime, ind_plus, which="plus")
                ind_prime = ind_plus
            elif traj is not None:
                ind_plus, ind_minus = ind, ind + 1
                traj.add_sample(thetaprime, ind_minus, which="minus")
                ind_prime = ind_minus
        else:
            # Recursion: Implicitly build the height j-1 left and right subtrees.
            if v == 1:
                (
                    thetaminus,
                    rminus,
                    gradminus,
                    thetaplus,
                    rplus,
                    gradplus,
                    thetaprime,
                    gradprime,
                    logpprime,
                    nprime,
                    sprime,
                    alphaprime,
                    nalphaprime,
                    ind_plus,
                    ind_minus,
                    ind_prime,
                ) = self.build_tree(theta, r, grad, logu, v, j - 1, epsilon, joint0, ind, traj, force_trajlen)
            else:
                (
                    thetaminus,
                    rminus,
                    gradminus,
                    thetaplus,
                    rplus,
                    gradplus,
                    thetaprime,
                    gradprime,
                    logpprime,
                    nprime,
                    sprime,
                    alphaprime,
                    nalphaprime,
                    ind_plus,
                    ind_minus,
                    ind_prime,
                ) = self.build_tree(theta, r, grad, logu, v, j - 1, epsilon, joint0, ind, traj, force_trajlen)

            # No need to keep going if the stopping criteria were met in the first subtree.
            if sprime == 1:
                if v == -1:
                    (
                        thetaminus,
                        rminus,
                        gradminus,
                        _,
                        _,
                        _,
                        thetaprime2,
                        gradprime2,
                        logpprime2,
                        nprime2,
                        sprime2,
                        alphaprime2,
                        nalphaprime2,
                        ind_plus,
                        ind_minus,
                        ind_prime2,
                    ) = self.build_tree(
                        thetaminus, rminus, gradminus, logu, v, j - 1, epsilon, joint0, ind_minus, traj, force_trajlen
                    )
                else:
                    (
                        _,
                        _,
                        _,
                        thetaplus,
                        rplus,
                        gradplus,
                        thetaprime2,
                        gradprime2,
                        logpprime2,
                        nprime2,
                        sprime2,
                        alphaprime2,
                        nalphaprime2,
                        ind_plus,
                        ind_minus,
                        ind_prime2,
                    ) = self.build_tree(
                        thetaplus, rplus, gradplus, logu, v, j - 1, epsilon, joint0, ind_plus, traj, force_trajlen
                    )
                # Choose which subtree to propagate a sample up from.
                if np.random.uniform() < (float(nprime2) / max(float(int(nprime) + int(nprime2)), 1.0)):
                    thetaprime = thetaprime2[:]
                    gradprime = gradprime2[:]
                    logpprime = logpprime2
                    ind_prime = ind_prime2

                # Update the number of valid points.
                nprime = int(nprime) + int(nprime2)
                # Update the stopping criterion.
                sprime = int(
                    sprime
                    and sprime2
                    and self.stop_criterion(
                        thetaminus, thetaplus, rminus, rplus, force_trajlen, max(ind_plus, ind_minus)
                    )
                )
                # Update the acceptance probability statistics.
                alphaprime = alphaprime + alphaprime2
                nalphaprime = nalphaprime + nalphaprime2

        return (
            thetaminus,
            rminus,
            gradminus,
            thetaplus,
            rplus,
            gradplus,
            thetaprime,
            gradprime,
            logpprime,
            nprime,
            sprime,
            alphaprime,
            nalphaprime,
            ind_plus,
            ind_minus,
            ind_prime,
        )

    def __call__(self, x, iter, beta):
        """Take one HMC trajectory step"""
        super(NUTSJump, self).__call__(x, iter, beta)

        x = np.atleast_1d(x)
        if len(np.shape(x)) > 1:
            raise ValueError("x is expected to be a 1-D array")

        q = self.forward(x)

        self.beta = beta

        # Always start evaluating the distribution and gradient.
        # Potential speed-up: obtain these values from elsewhere, since we must
        # have evaluated them already?
        logp, grad = self.func_grad_white(q)

        if self.epsilon is None and self.force_epsilon is None:
            # First time doing an HMC jump
            self.epsilon = self.find_reasonable_epsilon(q, grad, logp)
            # print("Find reasonable epsilon: ", self.epsilon)
            self.mu = np.log(10.0 * self.epsilon)  # For dual averaging
        elif self.epsilon is None and self.force_epsilon is not None:
            # Force epsilon to be a certain value
            self.epsilon = self.force_epsilon
            self.mu = np.log(10.0 * self.epsilon)  # For dual averaging
        elif self.force_epsilon is not None:
            # Force epsilon to be a certain value
            self.epsilon = self.force_epsilon

        # Update the mass matrix when in burn-in stage
        # Currently, there is a problem with adjusting the mass matrix.
        # Stepsize and mass matrix need to be tuned together?
        # if iter <= self.nburn:
        #    self.update_cf()

        # Set the start of the trajectory
        r0 = self.draw_momenta()
        joint = self.loghamiltonian(logp, r0)

        # Initial slice sampling variable
        logu = float(joint - np.random.exponential(1, size=1))

        # Initialize the binary tree for this trajectory
        sample = np.copy(q)
        lnprob = np.copy(logp)
        thetaminus = np.copy(sample)
        thetaplus = np.copy(sample)
        rminus = np.copy(r0)
        rplus = np.copy(r0)
        gradminus = np.copy(grad)
        gradplus = np.copy(grad)

        j = 0  # Initial tree heigth j = 0
        n = 1  # Initially, the only valid point is the initial point
        s = 1  # Stopping criterion

        # Reset the trajectory buffer
        self.traj.reset()
        self.traj.add_sample(thetaminus, self.traj.length())
        trajind, trajind_minus, trajind_plus, trajind_prime = 0, 0, 0, 0

        while s == 1:
            # Choose a direction. -1 = backwards, 1 = forwards.
            v = int(2 * (np.random.uniform() < 0.5) - 1)

            # Double the size of the tree.
            if v == -1:
                (
                    thetaminus,
                    rminus,
                    gradminus,
                    _,
                    _,
                    _,
                    thetaprime,
                    gradprime,
                    logpprime,
                    nprime,
                    sprime,
                    alpha,
                    nalpha,
                    trajind_plus,
                    trajind_minus,
                    trajind_prime,
                ) = self.build_tree(
                    thetaminus,
                    rminus,
                    gradminus,
                    logu,
                    v,
                    j,
                    self.epsilon,
                    joint,
                    trajind_minus,
                    self.traj,
                    self.force_trajlen,
                )
            else:
                (
                    _,
                    _,
                    _,
                    thetaplus,
                    rplus,
                    gradplus,
                    thetaprime,
                    gradprime,
                    logpprime,
                    nprime,
                    sprime,
                    alpha,
                    nalpha,
                    trajind_plus,
                    trajind_minus,
                    trajind_prime,
                ) = self.build_tree(
                    thetaplus,
                    rplus,
                    gradplus,
                    logu,
                    v,
                    j,
                    self.epsilon,
                    joint,
                    trajind_plus,
                    self.traj,
                    self.force_trajlen,
                )

            # Use Metropolis-Hastings to decide whether or not to move to a
            # point from the half-tree we just generated.
            _tmp = min(1, float(nprime) / float(n))
            if (sprime == 1) and (np.random.uniform() < _tmp):
                sample[:] = thetaprime[:]
                lnprob = np.copy(logpprime)
                grad = np.copy(gradprime)
                trajind = trajind_prime

            # Update number of valid points we've seen.
            n += nprime
            # Decide if it's time to stop.
            s = sprime and self.stop_criterion(
                thetaminus, thetaplus, rminus, rplus, self.force_trajlen, max(trajind_plus, trajind_minus)
            )
            # Increment depth.
            j += 1

            sys.stdout.flush()

        # Do adaptation of epsilon if we're still doing burn-in.
        if self.force_epsilon is None:
            eta = 1.0 / float(self.iter + self.t0)
            self.Hbar = (1.0 - eta) * self.Hbar + eta * (self.delta - alpha / float(nalpha))

            if iter <= self.nburn:
                # Still in the burn-in phase. So adjust epsilon
                self.epsilon = np.exp(self.mu - np.sqrt(self.iter) / self.gamma * self.Hbar)
                eta = self.iter**-self.kappa
                self.epsilonbar = np.exp((1.0 - eta) * np.log(self.epsilonbar) + eta * np.log(self.epsilon))

            else:
                self.epsilon = self.epsilonbar

        if self.trajectoryDir is not None:
            # Write the whole trajectory to file
            if iter <= self.nburn and self.write_burnin:
                trajfile_plus = os.path.join(self.trajectoryDir, "burnin-plus-{num:06d}.txt".format(num=iter))
                trajfile_minus = os.path.join(self.trajectoryDir, "burnin-minus-{num:06d}.txt".format(num=iter))
                trajfile_used = os.path.join(self.trajectoryDir, "burnin-used-{num:06d}.txt".format(num=iter))

                np.savetxt(trajfile_plus, self.traj.get_trajectory(which="plus")[0])
                np.savetxt(trajfile_minus, self.traj.get_trajectory(which="minus")[0])
                np.savetxt(trajfile_used, self.traj.get_used_trajectory(trajind))
            elif iter > self.nburn:
                trajfile_plus = os.path.join(self.trajectoryDir, "plus-{num:06d}.txt".format(num=iter - self.nburn))
                trajfile_minus = os.path.join(self.trajectoryDir, "minus-{num:06d}.txt".format(num=iter - self.nburn))
                trajfile_used = os.path.join(self.trajectoryDir, "used-{num:06d}.txt".format(num=iter - self.nburn))

                np.savetxt(trajfile_plus, self.traj.get_trajectory(which="plus")[0])
                np.savetxt(trajfile_minus, self.traj.get_trajectory(which="minus")[0])
                np.savetxt(trajfile_used, self.traj.get_used_trajectory(trajind))

        # We need to always accept this step, so the qxy is just the inverse MH ratio
        qxy = logp - lnprob

        return self.backward(sample), qxy



import os
import sys
import time

import numpy as np

from .nutsjump import HMCJump, MALAJump, NUTSJump

try:
    from mpi4py import MPI
except ImportError:
    print("Optional mpi4py package is not installed.  MPI support is not available.")
    from . import nompi4py as MPI

try:
    import acor
except ImportError:
    print(
        "Optional acor package is not installed. Acor is optionally used to calculate the "
        "effective chain length for output in the chain file."
    )
    pass


def shift_array(arr, num, fill_value=0.0):
    result = np.empty_like(arr)
    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr
    return result


class PTSampler(object):

    """
    Parallel Tempering Markov Chain Monte-Carlo (PTMCMC) sampler.
    This implementation uses an adaptive jump proposal scheme
    by default using both standard and single component Adaptive
    Metropolis (AM) and Differential Evolution (DE) jumps.

    This implementation also makes use of MPI (mpi4py) to run
    the parallel chains.

    Along with the AM and DE jumps, the user can add custom
    jump proposals with the ``addProposalToCycle`` fuction.

    @param ndim: number of dimensions in problem
    @param logl: log-likelihood function
    @param logp: log prior function (must be normalized for evidence evaluation)
    @param cov: Initial covariance matrix of model parameters for jump proposals
    @param covinds: Indices of parameters for which to perform adaptive jumps
    @param loglargs: any additional arguments (apart from the parameter vector) for
    log likelihood
    @param loglkwargs: any additional keyword arguments (apart from the parameter vector)
    for log likelihood
    @param logpargs: any additional arguments (apart from the parameter vector) for
    log like prior
    @param logl_grad: log-likelihood function, including gradients
    @param logp_grad: prior function, including gradients
    @param logpkwargs: any additional keyword arguments (apart from the parameter vector)
    for log prior
    @param outDir: Full path to output directory for chain files (default = ./chains)
    @param verbose: Update current run-status to the screen (default=True)
    @param resume: Resume from a previous chain (still in testing so beware) (default=False)

    """

    def __init__(
        self,
        ndim,
        logl,
        logp,
        cov,
        groups=None,
        loglargs=[],
        loglkwargs={},
        logpargs=[],
        logpkwargs={},
        logl_grad=None,
        logp_grad=None,
        comm=MPI.COMM_WORLD,
        outDir="./chains",
        verbose=True,
        resume=False,
        seed=None,
    ):

        # MPI initialization
        self.comm = comm
        self.MPIrank = self.comm.Get_rank()
        self.nchain = self.comm.Get_size()

        if self.MPIrank == 0:
            ss = np.random.SeedSequence(seed)
            child_seeds = ss.generate_state(self.nchain)
            self.stream = [np.random.default_rng(s) for s in child_seeds]
        else:
            self.stream = None
        self.stream = self.comm.scatter(self.stream, root=0)

        self.ndim = ndim
        self.logl = _function_wrapper(logl, loglargs, loglkwargs)
        self.logp = _function_wrapper(logp, logpargs, logpkwargs)
        if logl_grad is not None and logp_grad is not None:
            self.logl_grad = _function_wrapper(logl_grad, loglargs, loglkwargs)
            self.logp_grad = _function_wrapper(logp_grad, logpargs, logpkwargs)
        else:
            self.logl_grad = None
            self.logp_grad = None

        self.outDir = outDir
        self.verbose = verbose
        self.resume = resume

        # setup output file
        if not os.path.exists(self.outDir):
            try:
                os.makedirs(self.outDir)
            except OSError:
                pass

        # find indices for which to perform adaptive jumps
        self.groups = groups
        if groups is None:
            self.groups = [np.arange(0, self.ndim)]

        # set up covariance matrix
        self.cov = cov
        self.U = [[]] * len(self.groups)
        self.S = [[]] * len(self.groups)

        # do svd on parameter groups
        for ct, group in enumerate(self.groups):
            covgroup = np.zeros((len(group), len(group)))
            for ii in range(len(group)):
                for jj in range(len(group)):
                    covgroup[ii, jj] = self.cov[group[ii], group[jj]]

            self.U[ct], self.S[ct], v = np.linalg.svd(covgroup)

        self.M2 = np.zeros((ndim, ndim))
        self.mu = np.zeros(ndim)

        # initialize proposal cycle
        self.propCycle = []
        self.jumpDict = {}

        # indicator for auxilary jumps
        self.aux = []

    def initialize(
        self,
        Niter,
        ladder=None,
        Tmin=1,
        Tmax=None,
        Tskip=100,
        isave=1000,
        covUpdate=1000,
        SCAMweight=30,
        AMweight=20,
        DEweight=50,
        NUTSweight=20,
        HMCweight=20,
        MALAweight=0,
        burn=50000,
        HMCstepsize=0.1,
        HMCsteps=300,
        maxIter=None,
        thin=10,
        i0=0,
        neff=100000,
        writeHotChains=False,
        hotChain=False,
    ):
        """
        Initialize MCMC quantities

        @param maxIter: maximum number of iterations
        @Tmin: minumum temperature to use in temperature ladder

        """
        # get maximum number of iteration
        if maxIter is None and self.MPIrank > 0:
            maxIter = Niter
        elif maxIter is None and self.MPIrank == 0:
            maxIter = Niter

        self.ladder = ladder
        self.covUpdate = covUpdate
        self.SCAMweight = SCAMweight
        self.AMweight = AMweight
        self.DEweight = DEweight
        self.burn = burn
        self.Tskip = Tskip
        self.thin = thin
        self.isave = isave
        self.Niter = Niter
        self.neff = neff
        self.tstart = 0

        N = int(maxIter / thin)

        self._lnprob = np.zeros(N)
        self._lnlike = np.zeros(N)
        self._chain = np.zeros((N, self.ndim))
        self.naccepted = 0
        self.swapProposed = 0
        self.nswap_accepted = 0

        # set up covariance matrix and DE buffers
        if self.MPIrank == 0:
            self._AMbuffer = np.zeros((self.covUpdate, self.ndim))
            self._DEbuffer = np.zeros((self.burn, self.ndim))

        # ##### setup default jump proposal distributions ##### #

        # Gradient-based jumps
        if self.logl_grad is not None and self.logp_grad is not None:
            # DOES MALA do anything with the burnin? (Not adaptive enabled yet)
            malajump = MALAJump(self.logl_grad, self.logp_grad, self.cov, self.burn)
            self.addProposalToCycle(malajump, MALAweight)
            if MALAweight > 0:
                print("WARNING: MALA jumps are not working properly yet")

            # Perhaps have an option to adaptively tune the mass matrix?
            # Now that is done by defaulk
            hmcjump = HMCJump(
                self.logl_grad,
                self.logp_grad,
                self.cov,
                self.burn,
                stepsize=HMCstepsize,
                nminsteps=2,
                nmaxsteps=HMCsteps,
            )
            self.addProposalToCycle(hmcjump, HMCweight)

            # Target acceptance rate (delta) should be optimal for 0.6
            nutsjump = NUTSJump(
                self.logl_grad,
                self.logp_grad,
                self.cov,
                self.burn,
                trajectoryDir=None,
                write_burnin=False,
                force_trajlen=None,
                force_epsilon=None,
                delta=0.6,
            )
            self.addProposalToCycle(nutsjump, NUTSweight)

        # add SCAM
        self.addProposalToCycle(self.covarianceJumpProposalSCAM, self.SCAMweight)

        # add AM
        self.addProposalToCycle(self.covarianceJumpProposalAM, self.AMweight)

        # check length of jump cycle
        if len(self.propCycle) == 0:
            raise ValueError("No jump proposals specified!")

        # randomize cycle
        self.randomizeProposalCycle()

        # setup default temperature ladder
        if self.ladder is None:
            self.ladder = self.temperatureLadder(Tmin, Tmax=Tmax)

        # temperature for current chain
        self.temp = self.ladder[self.MPIrank]

        # hot chain sampling from prior
        if hotChain and self.MPIrank == self.nchain - 1:
            self.temp = 1e80
            self.fname = self.outDir + "/chain_hot.txt"
        else:
            self.fname = self.outDir + "/chain_{0}.txt".format(self.temp)

        # write hot chains
        self.writeHotChains = writeHotChains

        self.resumeLength = 0
        if self.resume and os.path.isfile(self.fname):
            if self.verbose:
                print("Resuming run from chain file {0}".format(self.fname))
            try:
                self.resumechain = np.loadtxt(self.fname)
                self.resumeLength = self.resumechain.shape[0]
            except ValueError:
                print("WARNING: Cant read in file. Removing last line.")
                os.system("sed -ie '$d' {0}".format(self.fname))
                self.resumechain = np.loadtxt(self.fname)
                self.resumeLength = self.resumechain.shape[0]
            self._chainfile = open(self.fname, "a")
        else:
            self._chainfile = open(self.fname, "w")
        self._chainfile.close()

    def updateChains(self, p0, lnlike0, lnprob0, iter):
        """
        Update chains after jump proposals

        """
        # update buffer
        if self.MPIrank == 0:
            self._AMbuffer[iter % self.covUpdate, :] = p0

        # put results into arrays
        if iter % self.thin == 0:
            ind = int(iter / self.thin)
            self._chain[ind, :] = p0
            self._lnlike[ind] = lnlike0
            self._lnprob[ind] = lnprob0

        # write to file
        if iter % self.isave == 0 and iter > 1 and iter > self.resumeLength:
            if self.writeHotChains or self.MPIrank == 0:
                self._writeToFile(iter)

            # write output covariance matrix
            np.save(self.outDir + "/cov.npy", self.cov)
            if self.MPIrank == 0 and self.verbose and iter > 1:
                sys.stdout.write("\r")
                sys.stdout.write(
                    "Finished %2.2f percent in %f s Acceptance rate = %g"
                    % (iter / self.Niter * 100, time.time() - self.tstart, self.naccepted / iter)
                )
                sys.stdout.flush()

    def sample(
        self,
        p0,
        Niter,
        ladder=None,
        Tmin=1,
        Tmax=None,
        Tskip=100,
        isave=1000,
        covUpdate=1000,
        SCAMweight=20,
        AMweight=20,
        DEweight=20,
        NUTSweight=20,
        MALAweight=20,
        HMCweight=20,
        burn=10000,
        HMCstepsize=0.1,
        HMCsteps=300,
        maxIter=None,
        thin=10,
        i0=0,
        neff=100000,
        writeHotChains=False,
        hotChain=False,
    ):
        """
        Function to carry out PTMCMC sampling.

        @param p0: Initial parameter vector
        @param self.Niter: Number of iterations to use for T = 1 chain
        @param ladder: User defined temperature ladder
        @param Tmin: Minimum temperature in ladder (default=1)
        @param Tmax: Maximum temperature in ladder (default=None)
        @param Tskip: Number of steps between proposed temperature swaps (default=100)
        @param isave: Number of iterations before writing to file (default=1000)
        @param covUpdate: Number of iterations between AM covariance updates (default=1000)
        @param SCAMweight: Weight of SCAM jumps in overall jump cycle (default=20)
        @param AMweight: Weight of AM jumps in overall jump cycle (default=20)
        @param DEweight: Weight of DE jumps in overall jump cycle (default=20)
        @param NUTSweight: Weight of the NUTS jumps in jump cycle (default=20)
        @param MALAweight: Weight of the MALA jumps in jump cycle (default=20)
        @param HMCweight: Weight of the HMC jumps in jump cycle (default=20)
        @param HMCstepsize: Step-size of the HMC jumps (default=0.1)
        @param HMCsteps: Maximum number of steps in an HMC trajectory (default=300)
        @param burn: Burn in time (DE jumps added after this iteration) (default=10000)
        @param maxIter: Maximum number of iterations for high temperature chains
                        (default=2*self.Niter)
        @param self.thin: Save every self.thin MCMC samples
        @param i0: Iteration to start MCMC (if i0 !=0, do not re-initialize)
        @param neff: Number of effective samples to collect before terminating

        """

        # get maximum number of iteration
        if maxIter is None and self.MPIrank > 0:
            maxIter = Niter
        elif maxIter is None and self.MPIrank == 0:
            maxIter = Niter

        # set up arrays to store lnprob, lnlike and chain
        # if picking up from previous run, don't re-initialize
        if i0 == 0:
            self.initialize(
                Niter,
                ladder=ladder,
                Tmin=Tmin,
                Tmax=Tmax,
                Tskip=Tskip,
                isave=isave,
                covUpdate=covUpdate,
                SCAMweight=SCAMweight,
                AMweight=AMweight,
                DEweight=DEweight,
                NUTSweight=NUTSweight,
                MALAweight=MALAweight,
                HMCweight=HMCweight,
                burn=burn,
                HMCstepsize=HMCstepsize,
                HMCsteps=HMCsteps,
                maxIter=maxIter,
                thin=thin,
                i0=i0,
                neff=neff,
                writeHotChains=writeHotChains,
                hotChain=hotChain,
            )

        # compute lnprob for initial point in chain

        # if resuming, just start with first point in chain
        if self.resume and self.resumeLength > 0:
            p0, lnlike0, lnprob0 = self.resumechain[0, :-4], self.resumechain[0, -3], self.resumechain[0, -4]
        else:
            # compute prior
            lp = self.logp(p0)

            if lp == float(-np.inf):

                lnprob0 = -np.inf
                lnlike0 = -np.inf

            else:

                lnlike0 = self.logl(p0)
                lnprob0 = 1 / self.temp * lnlike0 + lp

        # record first values
        self.updateChains(p0, lnlike0, lnprob0, i0)

        self.comm.barrier()

        # start iterations
        iter = i0
        self.tstart = time.time()
        runComplete = False
        Neff = 0
        while runComplete is False:
            iter += 1
            self.comm.barrier()  # make sure all processes are at the same iteration
            # call PTMCMCOneStep
            p0, lnlike0, lnprob0 = self.PTMCMCOneStep(p0, lnlike0, lnprob0, iter)

            # compute effective number of samples
            if iter % 1000 == 0 and iter > 2 * self.burn and self.MPIrank == 0:
                try:
                    Neff = iter / max(
                        1,
                        np.nanmax([acor.acor(self._chain[self.burn : (iter - 1), ii])[0] for ii in range(self.ndim)]),
                    )
                    # print('\n {0} effective samples'.format(Neff))
                except NameError:
                    Neff = 0
                    pass

            # stop if reached maximum number of iterations
            if self.MPIrank == 0 and iter >= self.Niter - 1:
                if self.verbose:
                    print("\nRun Complete")
                runComplete = True

            # stop if reached effective number of samples
            if self.MPIrank == 0 and int(Neff) > self.neff:
                if self.verbose:
                    print("\nRun Complete with {0} effective samples".format(int(Neff)))
                runComplete = True

            runComplete = self.comm.bcast(runComplete, root=0)

    def PTMCMCOneStep(self, p0, lnlike0, lnprob0, iter):
        """
        Function to carry out PTMCMC sampling.

        @param p0: Initial parameter vector
        @param lnlike0: Initial log-likelihood value
        @param lnprob0: Initial log probability value
        @param iter: iteration number

        @return p0: next value of parameter vector after one MCMC step
        @return lnlike0: next value of likelihood after one MCMC step
        @return lnprob0: next value of posterior after one MCMC step

        """
        # update covariance matrix
        if (iter - 1) % self.covUpdate == 0 and (iter - 1) != 0 and self.MPIrank == 0:
            self._updateRecursive(iter - 1, self.covUpdate)

            # broadcast to other chains
            [self.comm.send(self.cov, dest=rank + 1, tag=111) for rank in range(self.nchain - 1)]

        # update covariance matrix
        if (iter - 1) % self.covUpdate == 0 and (iter - 1) != 0 and self.MPIrank > 0:
            self.cov[:, :] = self.comm.recv(source=0, tag=111)
            for ct, group in enumerate(self.groups):
                covgroup = np.zeros((len(group), len(group)))
                for ii in range(len(group)):
                    for jj in range(len(group)):
                        covgroup[ii, jj] = self.cov[group[ii], group[jj]]

                self.U[ct], self.S[ct], v = np.linalg.svd(covgroup)

        # update DE buffer
        if (iter - 1) % self.burn == 0 and (iter - 1) != 0 and self.MPIrank == 0:
            self._updateDEbuffer(iter - 1, self.burn)

            # broadcast to other chains
            [self.comm.send(self._DEbuffer, dest=rank + 1, tag=222) for rank in range(self.nchain - 1)]

        # update DE buffer
        if (iter - 1) % self.burn == 0 and (iter - 1) != 0 and self.MPIrank > 0:
            self._DEbuffer = self.comm.recv(source=0, tag=222)

            # randomize cycle
            if self.DEJump not in self.propCycle:
                self.addProposalToCycle(self.DEJump, self.DEweight)
                self.randomizeProposalCycle()

        # after burn in, add DE jumps
        if (iter - 1) == self.burn and self.MPIrank == 0:
            if self.verbose:
                print("Adding DE jump with weight {0}".format(self.DEweight))
            self.addProposalToCycle(self.DEJump, self.DEweight)

            # randomize cycle
            self.randomizeProposalCycle()

        # jump proposal ###

        # if resuming, just use previous chain points
        if self.resume and self.resumeLength > 0 and iter < self.resumeLength:
            p0, lnlike0, lnprob0 = self.resumechain[iter, :-4], self.resumechain[iter, -3], self.resumechain[iter, -4]

            # update acceptance counter
            self.naccepted = iter * self.resumechain[iter, -2]
        else:
            y, qxy, jump_name = self._jump(p0, iter)
            self.jumpDict[jump_name][0] += 1

            # compute prior and likelihood
            lp = self.logp(y)

            if lp == -np.inf:

                newlnprob = -np.inf

            else:

                newlnlike = self.logl(y)
                newlnprob = 1 / self.temp * newlnlike + lp

            # hastings step
            diff = newlnprob - lnprob0 + qxy
            if diff > np.log(self.stream.random()):

                # accept jump
                p0, lnlike0, lnprob0 = y, newlnlike, newlnprob

                # update acceptance counter
                self.naccepted += 1
                self.jumpDict[jump_name][1] += 1
        # temperature swap
        if iter % self.Tskip == 0 and self.nchain > 1:
            p0, lnlike0, lnprob0 = self.PTswap(p0, lnlike0, lnprob0, iter)

        self.updateChains(p0, lnlike0, lnprob0, iter)

        return p0, lnlike0, lnprob0

    def PTswap(self, p0, lnlike0, lnprob0, iter):
        """
        Do parallel tempering swap.

        @param p0: current parameter vector
        @param lnlike0: current log-likelihood
        @param lnprob0: current log posterior value
        @param iter: current iteration number

        @return swapReturn: 0 = no swap proposed,
        1 = swap proposed and rejected,
        2 = swap proposed and accepted

        @return p0: new parameter vector
        @return lnlike0: new log-likelihood
        @return lnprob0: new log posterior value

        Repurposed from Neil Cornish/Bence Becsy's code:
        """
        Ts = self.ladder

        log_Ls = self.comm.gather(lnlike0, root=0)  # list of likelihoods from each chain
        p0s = self.comm.gather(p0, root=0)  # list of parameter arrays from each chain

        if self.MPIrank == 0:
            # set up map to help keep track of swaps
            swap_map = list(range(self.nchain))

            # loop through and propose a swap at each chain (starting from hottest chain and going down in T)
            # and keep track of results in swap_map
            for swap_chain in reversed(range(self.nchain - 1)):
                log_acc_ratio = -log_Ls[swap_map[swap_chain]] / Ts[swap_chain]
                log_acc_ratio += -log_Ls[swap_map[swap_chain + 1]] / Ts[swap_chain + 1]
                log_acc_ratio += log_Ls[swap_map[swap_chain + 1]] / Ts[swap_chain]
                log_acc_ratio += log_Ls[swap_map[swap_chain]] / Ts[swap_chain + 1]

                acc_ratio = np.exp(log_acc_ratio)
                if self.stream.uniform() <= acc_ratio:
                    swap_map[swap_chain], swap_map[swap_chain + 1] = swap_map[swap_chain + 1], swap_map[swap_chain]
                    self.nswap_accepted += 1
                    self.swapProposed += 1
                else:
                    self.swapProposed += 1

            # loop through the chains and record the new samples and log_Ls
            for j in range(self.nchain):
                p0s[j] = p0s[swap_map[j]]
                log_Ls[j] = log_Ls[swap_map[j]]

        # broadcast the new samples and log_Ls to all chains
        p0 = self.comm.scatter(p0s, root=0)
        lnlike0 = self.comm.scatter(log_Ls, root=0)

        # calculate new posterior values
        lnprob0 = 1 / self.temp * lnlike0 + self.logp(p0)

        return p0, lnlike0, lnprob0

    def temperatureLadder(self, Tmin, Tmax=None, tstep=None):
        """
        Method to compute temperature ladder. At the moment this uses
        a geometrically spaced temperature ladder with a temperature
        spacing designed to give 25 % temperature swap acceptance rate.

        """

        # TODO: make options to do other temperature ladders

        if self.nchain > 1:
            if tstep is None and Tmax is None:
                tstep = 1 + np.sqrt(2 / self.ndim)
            elif tstep is None and Tmax is not None:
                tstep = np.exp(np.log(Tmax / Tmin) / (self.nchain - 1))
            ladder = np.zeros(self.nchain)
            for ii in range(self.nchain):
                ladder[ii] = Tmin * tstep**ii
        else:
            ladder = np.array([1])

        return ladder

    def _writeToFile(self, iter):
        """
        Function to write chain file. File has 3+ndim columns,
        the first is log-posterior (unweighted), log-likelihood,
        and acceptance probability, followed by parameter values.

        @param iter: Iteration of sampler

        """

        self._chainfile = open(self.fname, "a+")
        for jj in range((iter - self.isave), iter, self.thin):
            ind = int(jj / self.thin)
            pt_acc = 1
            if self.MPIrank < self.nchain - 1 and self.swapProposed != 0:
                pt_acc = self.nswap_accepted / self.swapProposed

            self._chainfile.write("\t".join(["%22.22f" % (self._chain[ind, kk]) for kk in range(self.ndim)]))
            self._chainfile.write(
                "\t%f\t%f\t%f\t%f\n" % (self._lnprob[ind], self._lnlike[ind], self.naccepted / iter, pt_acc)
            )
        self._chainfile.close()

        # write jump statistics files ####

        # only for T=1 chain
        if self.MPIrank == 0:

            # first write file contaning jump names and jump rates
            fout = open(self.outDir + "/jumps.txt", "w")
            njumps = len(self.propCycle)
            ujumps = np.array(list(set(self.propCycle)))
            for jump in ujumps:
                fout.write("%s %4.2g\n" % (jump.__name__, np.sum(np.array(self.propCycle) == jump) / njumps))

            fout.close()

            # now write jump statistics for each jump proposal
            for jump in self.jumpDict:
                fout = open(self.outDir + "/" + jump + "_jump.txt", "a+")
                fout.write("%g\n" % (self.jumpDict[jump][1] / max(1, self.jumpDict[jump][0])))
                fout.close()

    # function to update covariance matrix for jump proposals
    def _updateRecursive(self, iter, mem):
        """
        Function to recursively update sample covariance matrix.

        @param iter: Iteration of sampler
        @param mem: Number of steps between updates

        """

        it = iter - mem
        ndim = self.ndim

        if it == 0:
            self.M2 = np.zeros((ndim, ndim))
            self.mu = np.zeros(ndim)

        for ii in range(mem):
            diff = np.zeros(ndim)
            it += 1
            for jj in range(ndim):

                diff[jj] = self._AMbuffer[ii, jj] - self.mu[jj]
                self.mu[jj] += diff[jj] / it

            self.M2 += np.outer(diff, (self._AMbuffer[ii, :] - self.mu))

        self.cov[:, :] = self.M2 / (it - 1)

        # do svd on parameter groups
        for ct, group in enumerate(self.groups):
            covgroup = np.zeros((len(group), len(group)))
            for ii in range(len(group)):
                for jj in range(len(group)):
                    covgroup[ii, jj] = self.cov[group[ii], group[jj]]

            self.U[ct], self.S[ct], v = np.linalg.svd(covgroup)

    # update DE buffer samples
    def _updateDEbuffer(self, iter, burn):
        """
        Update Differential Evolution with last burn
        values in the total chain

        @param iter: Iteration of sampler
        @param burn: Total number of samples in DE buffer

        """

        self._DEbuffer = shift_array(self._DEbuffer, -len(self._AMbuffer))  # shift DEbuffer to the left
        self._DEbuffer[-len(self._AMbuffer) :] = self._AMbuffer  # add new samples to the new empty spaces

    # SCAM jump
    def covarianceJumpProposalSCAM(self, x, iter, beta):
        """
        Single Component Adaptive Jump Proposal. This function will occasionally
        jump in more than 1 parameter. It will also occasionally use different
        jump sizes to ensure proper mixing.

        @param x: Parameter vector at current position
        @param iter: Iteration of sampler
        @param beta: Inverse temperature of chain

        @return: q: New position in parameter space
        @return: qxy: Forward-Backward jump probability

        """

        q = x.copy()
        qxy = 0

        # choose group
        jumpind = self.stream.integers(0, len(self.groups))
        ndim = len(self.groups[jumpind])

        # adjust step size
        prob = self.stream.random()

        # large jump
        if prob > 0.97:
            scale = 10

        # small jump
        elif prob > 0.9:
            scale = 0.2

        # small-medium jump
        # elif prob > 0.6:

        # standard medium jump
        else:
            scale = 1.0

        # adjust scale based on temperature
        if self.temp <= 100:
            scale *= np.sqrt(self.temp)

        # get parmeters in new diagonalized basis
        # y = np.dot(self.U.T, x[self.covinds])

        # make correlated componentwise adaptive jump
        ind = np.unique(self.stream.integers(0, ndim, 1))
        neff = len(ind)
        cd = 2.4 / np.sqrt(2 * neff) * scale

        q[self.groups[jumpind]] += (
            self.stream.standard_normal() * cd * np.sqrt(self.S[jumpind][ind]) * self.U[jumpind][:, ind].flatten()
        )

        return q, qxy

    # AM jump
    def covarianceJumpProposalAM(self, x, iter, beta):
        """
        Adaptive Jump Proposal. This function will occasionally
        use different jump sizes to ensure proper mixing.

        @param x: Parameter vector at current position
        @param iter: Iteration of sampler
        @param beta: Inverse temperature of chain

        @return: q: New position in parameter space
        @return: qxy: Forward-Backward jump probability

        """

        q = x.copy()
        qxy = 0

        # choose group
        jumpind = self.stream.integers(0, len(self.groups))

        # adjust step size
        prob = self.stream.random()

        # large jump
        if prob > 0.97:
            scale = 10

        # small jump
        elif prob > 0.9:
            scale = 0.2

        # small-medium jump
        # elif prob > 0.6:
        #    scale = 0.5

        # standard medium jump
        else:
            scale = 1.0

        # adjust scale based on temperature
        if self.temp <= 100:
            scale *= np.sqrt(self.temp)

        # get parmeters in new diagonalized basis
        y = np.dot(self.U[jumpind].T, x[self.groups[jumpind]])

        # make correlated componentwise adaptive jump
        ind = np.arange(len(self.groups[jumpind]))
        neff = len(ind)
        cd = 2.4 / np.sqrt(2 * neff) * scale

        y[ind] = y[ind] + self.stream.standard_normal(neff) * cd * np.sqrt(self.S[jumpind][ind])
        q[self.groups[jumpind]] = np.dot(self.U[jumpind], y)

        return q, qxy

    # Differential evolution jump
    def DEJump(self, x, iter, beta):
        """
        Differential Evolution Jump. This function will  occasionally
        use different jump sizes to ensure proper mixing.

        @param x: Parameter vector at current position
        @param iter: Iteration of sampler
        @param beta: Inverse temperature of chain

        @return: q: New position in parameter space
        @return: qxy: Forward-Backward jump probability

        """

        # get old parameters
        q = x.copy()
        qxy = 0

        # choose group
        jumpind = self.stream.integers(0, len(self.groups))
        ndim = len(self.groups[jumpind])

        bufsize = len(self._DEbuffer)

        # draw a random integer from 0 - iter
        mm = self.stream.integers(0, bufsize)
        nn = self.stream.integers(0, bufsize)

        # make sure mm and nn are not the same iteration
        while mm == nn:
            nn = self.stream.integers(0, bufsize)

        # get jump scale size
        prob = self.stream.random()

        # mode jump
        if prob > 0.5:
            scale = 1.0

        else:
            scale = self.stream.random() * 2.4 / np.sqrt(2 * ndim) * np.sqrt(1 / beta)

        for ii in range(ndim):

            # jump size
            sigma = self._DEbuffer[mm, self.groups[jumpind][ii]] - self._DEbuffer[nn, self.groups[jumpind][ii]]

            # jump
            q[self.groups[jumpind][ii]] += scale * sigma

        return q, qxy

    # add jump proposal distribution functions
    def addProposalToCycle(self, func, weight):
        """
        Add jump proposal distributions to cycle with a given weight.

        @param func: jump proposal function
        @param weight: jump proposal function weight in cycle

        """

        # get length of cycle so far
        length = len(self.propCycle)

        # check for 0 weight
        if weight == 0:
            # print('ERROR: Can not have 0 weight in proposal cycle!')
            # sys.exit()
            return

        # add proposal to cycle
        for ii in range(length, length + weight):
            self.propCycle.append(func)

        # add to jump dictionary and initialize file
        if func.__name__ not in self.jumpDict:
            self.jumpDict[func.__name__] = [0, 0]
            fout = open(self.outDir + "/" + func.__name__ + "_jump.txt", "w")
            fout.close()

    # add auxilary jump proposal distribution functions
    def addAuxilaryJump(self, func):
        """
        Add auxilary jump proposal distribution. This will be called after every
        standard jump proposal. Examples include cyclic boundary conditions and
        pulsar phase fixes

        @param func: jump proposal function

        """

        # set auxilary jump
        self.aux.append(func)

    # randomized proposal cycle
    def randomizeProposalCycle(self):
        """
        Randomize proposal cycle that has already been filled

        """

        # get length of full cycle
        length = len(self.propCycle)

        # get random integers
        index = np.arange(length)
        self.stream.shuffle(index)

        # randomize proposal cycle
        self.randomizedPropCycle = [self.propCycle[ind] for ind in index]

    # call proposal functions from cycle
    def _jump(self, x, iter):
        """
        Call Jump proposals

        """

        # get length of cycle
        length = len(self.propCycle)

        # call function
        ind = self.stream.integers(0, length)
        q, qxy = self.propCycle[ind](x, iter, 1 / self.temp)

        # axuilary jump
        if len(self.aux) > 0:
            for aux in self.aux:
                q, qxy_aux = aux(x, q, iter, 1 / self.temp)
                qxy += qxy_aux

        return q, qxy, self.propCycle[ind].__name__

    # TODO: jump statistics


class _function_wrapper(object):

    """
    This is a hack to make the likelihood function pickleable when ``args``
    or ``kwargs`` are also included.

    """

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def __call__(self, x):
        return self.f(x, *self.args, **self.kwargs)



#!/usr/bin/env python
# coding: utf-8


from pathlib import Path

import numpy as np

from PTMCMCSampler import PTMCMCSampler

# ## Define the likelihood and posterior
#
# Functions must read in parameter vector and output log-likelihood or log-prior.
# Usually easiest to use a class if you need to store some other data or parameters


class GaussianLikelihood(object):
    def __init__(self, ndim=2, pmin=-10, pmax=10):

        self.a = np.ones(ndim) * pmin
        self.b = np.ones(ndim) * pmax

        # get means
        self.mu = np.random.uniform(pmin, pmax, ndim)

        # ... and a positive definite, non-trivial covariance matrix.
        cov = 0.5 - np.random.rand(ndim**2).reshape((ndim, ndim))
        cov = np.triu(cov)
        cov += cov.T - np.diag(cov.diagonal())
        self.cov = np.dot(cov, cov)

        # Invert the covariance matrix first.
        self.icov = np.linalg.inv(self.cov)

    def lnlikefn(self, x):
        diff = x - self.mu
        return -np.dot(diff, np.dot(self.icov, diff)) / 2.0

    def lnpriorfn(self, x):

        if np.all(self.a <= x) and np.all(self.b >= x):
            return 0.0
        else:
            return -np.inf


# ## Setup Gaussian model class

# In[3]:


ndim = 20
pmin, pmax = 0.0, 10.0
glo = GaussianLikelihood(ndim=ndim, pmin=pmin, pmax=pmax)


# ## Setup sampler
#
# Need to initalize the sample at ```p0``` and give an inital jump covariance matrix ```cov```.

# In[4]:


# Set the start position and the covariance
p0 = np.random.uniform(pmin, pmax, ndim)
cov = np.eye(ndim) * 0.1**2


# In[5]:


sampler = PTMCMCSampler.PTSampler(
    ndim, glo.lnlikefn, glo.lnpriorfn, np.copy(cov), outDir=str(Path(__file__).parent / "chains")
)


# ## Add custom jump
#
# Can add custom jump in the following way

# In[6]:


class UniformJump(object):
    def __init__(self, pmin, pmax):
        """Draw random parameters from pmin, pmax"""
        self.pmin = pmin
        self.pmax = pmax

    def jump(self, x, it, beta):
        """
        Function prototype must read in parameter vector x,
        sampler iteration number it, and inverse temperature beta
        """

        # log of forward-backward jump probability
        lqxy = 0

        # uniformly drawm parameters
        q = np.random.uniform(self.pmin, self.pmax, len(x))

        return q, lqxy


# In[7]:


# add to jump proposal cycle
ujump = UniformJump(pmin, pmax)
sampler.addProposalToCycle(ujump.jump, 5)


# ## Run Sampler for 100000 steps
#
# Different jump proposal weights are given as integers.
# For example we have used a weight of 20 for all three proposals here.
# That means that each will be used with a probability of 20/60 = 1/3.

# In[8]:


sampler.sample(p0, 10000, burn=500, thin=1, covUpdate=500, SCAMweight=20, AMweight=20, DEweight=20)



from setuptools import setup

setup(
    name="ptmcmcsampler",
    author="Justin A. Ellis",
    author_email="justin.ellis18@gmail.com",
    packages=["PTMCMCSampler"],
    package_dir={"PTMCMCSampler": "PTMCMCSampler"},
    url="https://github.com/jellis18/PTMCMCSampler",
    license="MIT",
    zip_safe=False,
    description="Parallel tempering MCMC sampler written in Python",
    long_description=open("README.md").read() + "\n\n" + "---------\n\n" + open("HISTORY.md").read(),
    long_description_content_type="text/markdown",
    package_data={"": ["README.md", "HISTORY.md"]},
    install_requires=["numpy>=1.16.3", "scipy>=1.2.0"],
    python_requires=">=3.7",
    extras_require={"mpi": ["mpi4py>=3.0.3"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import shutil
from unittest import TestCase

import numpy as np
import scipy.linalg as sl
import scipy.optimize as so
from mpi4py import MPI

from PTMCMCSampler import PTMCMCSampler
from PTMCMCSampler import nompi4py as MPIDUMMY


class GaussianLikelihood(object):
    def __init__(self, ndim=2, pmin=-10, pmax=10):

        self.a = np.ones(ndim) * pmin
        self.b = np.ones(ndim) * pmax

    def lnlikefn(self, x):
        return -0.5 * np.sum(x**2) - len(x) * 0.5 * np.log(2 * np.pi)

    def lnlikefn_grad(self, x):
        ll = -0.5 * np.sum(x**2) - len(x) * 0.5 * np.log(2 * np.pi)
        ll_grad = -x
        return ll, ll_grad

    def lnpriorfn(self, x):

        if np.all(self.a <= x) and np.all(self.b >= x):
            return 0.0
        else:
            return -np.inf
        return 0.0

    def lnpriorfn_grad(self, x):
        return self.lnpriorfn(x), np.zeros_like(x)

    def lnpost_grad(self, x):
        ll, ll_grad = self.lnlikefn_grad(x)
        lp, lp_grad = self.lnpriorfn_grad(x)
        return ll + lp, ll_grad + lp_grad

    def lnpost(self, x):
        return self.lnpost_grad(x)[0]

    def hessian(self, x):
        return -np.eye(len(x))


class intervalTransform(object):
    """
    Wrapper class of the likelihood for Hamiltonian samplers. This implements a
    coordinate transformation for all parameters from an interval to all real numbers.
    """

    def __init__(self, likob, pmin=None, pmax=None):
        """Initialize the intervalLikelihood with a ptaLikelihood object"""
        self.likob = likob

        if pmin is None:
            self.a = likob.a
        else:
            self.a = pmin * np.ones_like(likob.a)

        if pmax is None:
            self.b = likob.b
        else:
            self.b = pmax * np.ones_like(likob.b)

    def forward(self, x):
        """Forward transform the real coordinates (on the interval) to the
        transformed coordinates (on all real numbers)
        """
        p = np.atleast_2d(x.copy())
        posinf, neginf = (self.a == x), (self.b == x)
        m = ~(posinf | neginf)
        p[:, m] = np.log((p[:, m] - self.a[m]) / (self.b[m] - p[:, m]))
        p[:, posinf] = np.inf
        p[:, neginf] = -np.inf
        return p.reshape(x.shape)

    def backward(self, p):
        """Backward transform the transformed coordinates (on all real numbers)
        to the real coordinates (on the interval)
        """
        x = np.atleast_2d(p.copy())
        x[:, :] = (self.b[:] - self.a[:]) * np.exp(x[:, :]) / (1 + np.exp(x[:, :])) + self.a[:]
        return x.reshape(p.shape)

    def logjacobian_grad(self, p):
        """Return the log of the Jacobian at point p"""
        lj = np.sum(np.log(self.b[:] - self.a[:]) + p[:] - 2 * np.log(1.0 + np.exp(p[:])))

        lj_grad = np.zeros_like(p)
        lj_grad[:] = (1 - np.exp(p[:])) / (1 + np.exp(p[:]))
        return lj, lj_grad

    def dxdp(self, p):
        """Derivative of x wrt p (jacobian for chain-rule) - diagonal"""
        pp = np.atleast_2d(p)
        d = np.ones_like(pp)
        d[:, :] = (self.b[:] - self.a[:]) * np.exp(pp[:, :]) / (1 + np.exp(pp[:, :])) ** 2
        return d.reshape(p.shape)

    def d2xd2p(self, p):
        """Derivative of x wrt p (jacobian for chain-rule) - diagonal"""
        pp = np.atleast_2d(p)
        d = np.zeros_like(pp)
        d[:, :] = (self.b[:] - self.a[:]) * (np.exp(2 * pp[:, :]) - np.exp(pp[:, :])) / (1 + np.exp(pp[:, :])) ** 3
        return d.reshape(p.shape)

    def logjac_hessian(self, p):
        """The Hessian of the log-jacobian"""
        # p should not be more than one-dimensional
        assert len(p.shape) == 1

        return np.diag(-2 * np.exp(p) / (1 + np.exp(p)) ** 2)

    def lnlikefn_grad(self, p, **kwargs):
        """The log-likelihood in the new coordinates"""
        x = self.backward(p)
        ll, ll_grad = self.likob.lnlikefn_grad(x, **kwargs)
        lj, lj_grad = self.logjacobian_grad(p)
        return ll + lj, ll_grad * self.dxdp(p) + lj_grad

    def lnlikefn(self, p, **kwargs):
        return self.lnlikefn_grad(p)[0]

    def lnpriorfn_grad(self, p, **kwargs):
        """The log-prior in the new coordinates. Do not include the Jacobian"""
        x = self.backward(p)
        lp, lp_grad = self.likob.lnpriorfn_grad(x)
        return lp, lp_grad * self.dxdp(p)

    def lnpriorfn(self, p, **kwargs):
        return self.lnpriorfn_grad(p)[0]

    def logpostfn_grad(self, p, **kwargs):
        """The log-posterior in the new coordinates"""
        x = self.backward(p)
        lp, lp_grad = self.likob.lnpost_grad(x)
        lj, lj_grad = self.logjacobian_grad(p)
        return lp + lj, lp_grad * self.dxdp(p) + lj_grad

    def hessian(self, p):
        """The Hessian matrix in the new coordinates"""
        # p should not be more than one-dimensional
        assert len(p.shape) == 1

        # Get quantities from un-transformed distribution
        x = self.backward(p)
        orig_hessian = self.likob.hessian(x)
        _, orig_lp_grad = self.likob.lnpost_grad(x)

        # Transformation properties
        hessian = self.logjac_hessian(p)
        dxdpf = np.diag(self.dxdp(p))

        hessian += np.dot(dxdpf.T, np.dot(orig_hessian, dxdpf))
        hessian -= np.diag(self.d2xd2p(p) * orig_lp_grad)

        return hessian


class TestNuts(TestCase):
    @classmethod
    def tearDownClass(cls):
        shutil.rmtree("chains")

    def setUp(self) -> None:
        self.comm = MPI.COMM_WORLD

    def test_nuts(self):
        ndim = 40
        glo = GaussianLikelihood(ndim=ndim, pmin=0.0, pmax=10.0)
        glt = intervalTransform(glo, pmin=0.0, pmax=10)
        gl = glt

        p0 = np.ones(ndim) * 0.01

        # Maximize using scipy
        result = so.minimize(
            lambda x: -gl.lnlikefn(x),
            p0,
            jac=lambda x: -gl.lnlikefn_grad(x)[1],
            method="Newton-CG",
            hess=lambda x: -gl.hessian(x),
            options={"disp": True},
        )

        # Set the start position and the covariance
        p0 = result["x"]
        h0 = gl.hessian(p0)
        cov = sl.cho_solve(sl.cho_factor(-h0), np.eye(len(h0)))

        sampler = PTMCMCSampler.PTSampler(
            ndim,
            gl.lnlikefn,
            gl.lnpriorfn,
            np.copy(cov),
            logl_grad=gl.lnlikefn_grad,
            logp_grad=gl.lnpriorfn_grad,
            outDir="./chains",
            comm=self.comm,
        )

        sampler.sample(
            p0,
            1000,
            burn=500,
            thin=1,
            covUpdate=500,
            SCAMweight=10,
            AMweight=10,
            DEweight=10,
            NUTSweight=10,
            HMCweight=10,
            MALAweight=0,
            HMCsteps=100,
            HMCstepsize=0.4,
        )


class TestNutsNoMPI(TestNuts):
    def setUp(self) -> None:
        self.comm = MPIDUMMY.COMM_WORLD
